/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1997 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, modify, and distribute this material for any
purpose and without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies, and
that the name of ETL not be used in advertising or publicity pertaining
to this material without the specific, prior written permission of an
authorized representative of ETL.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	mkmake.c (Makefile preprocessor)
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:
History:
	970105	created
//////////////////////////////////////////////////////////////////////#*/
#include <stdio.h>
#include <string.h>
extern FILE *tmpfile();
extern char *getenv();

#define MKMKMK_EXE "mkmkmk.exe"
#define SRCFILE_LIST	"srcfiles"

#include <fcntl.h>
#ifdef O_BINARY
int _fmode;
static setBinaryIO(){ _fmode = O_BINARY; }
#else
static setBinaryIO(){ }
#endif

#include <sys/types.h>
#include <sys/stat.h>
#ifdef S_ISUID /* set UID on execution */
#define MaybeOnUnix	1
#else
#define MaybeOnUnix	0
#endif

#ifdef _MSC_VER
/*########################################### MSWIN ############*/
#define SYST		"MSWIN"
#define BESILENT	"-s -nologo"
#define CCOUT		"/Fo"
#include <process.h>
#define pipe(sv)	_pipe(sv,0x2000,O_BINARY)
#define popen(com,mod)	_popen(com,mod)
#define pclose(fp)	_pclose(fp)
#define sleep(s)	_sleep(s*1000)
#define QUOTE_ARG
#define RETRY_RENAME
#define WITH_SPAWN

#else /* !_MSC_VER */
#ifdef __EMX__
/*########################################### OS2EMX ###########*/
#define SYST		"OS2EMX"
#define BESILENT	"-s"
#define CCOUT		"-o "
#include <process.h>
#define WITH_SPAWN

#else /* !__EMX__ */
/*########################################### UNIX #############*/
#define SYST		"UNIX"
#define BESILENT	"-s"
#define CCOUT		"-o "
#include <sys/wait.h>
#ifndef WEXITSTATUS
#define WEXITSTATUS(status)	((status >> 8) & 0xFF)
#endif

#endif /* __EMX__ */
#endif /* _MSC_VER */

static char *GETCWD(cwd,size)
	char *cwd;
{
#ifdef MKMKMK
	fprintf(stderr,"#### getcwd() is not available in mkmkmk.\n");
	return NULL;
#else
	extern char *getcwd();
	return getcwd(cwd,size);
#endif
}

static char *STRSTR(s1,s2)
	char *s1,*s2;
{	int len;

	len = strlen(s2);
	for(; *s1; s1++ )
		if( *s1 == *s2 && strncmp(s1,s2,len) == 0 )
			return s1;
	return NULL;
}

static char *_pathsep;
static char *_PATHSEP(){
	char cwd[1024];
	if( _pathsep == NULL ){
		GETCWD(cwd,sizeof(cwd));
		if( strchr(cwd,'\\') )
			_pathsep = "\\";
		else	_pathsep = "/";
	}
	return _pathsep;
}
#define PATHSEP _PATHSEP()

static char *_exeext;
static char *_EXEEXT(){
}


static char *MAKEenv(oenv,nenv)
	char *oenv,*nenv;
{	char *rpath,apath[1024];

	if( strncmp(oenv,"MAKE=",5) != 0 )
		return oenv;
	rpath = &oenv[5];
	if( *rpath == 0 )
		rpath = "make";
	if( which(rpath,apath) && apath[0] == PATHSEP[0] )
		sprintf(nenv,"MAKE=%s",apath);
	else	sprintf(nenv,"MAKE=%s",rpath);
	return nenv;
}

#ifndef DEFCC
#define DEFCC "cc"
#endif

mkmkmk(ac,av)
	char *av[];
{	FILE *tmp;
	char *arg,*nav[256];
	char libdir[2048],hdrdir[2048];
	int ai,nac;

	if( ac < 4 || strcmp(av[1],"-mkmkmk") != 0 ){
		for( ai = 2; ai < ac; ai++ )
			printf("[%d] %s\n",ai,av[ai]);
		fprintf(stderr,"Usage: %s -mkmkmk CC mkmake.c\n",av[0]);
		exit(-1);
	}

	nac = 0;
	libdir[0] = 0;
	for( ai = 2; ai < ac; ai++ ){
		arg = nav[nac++] = av[ai];
		if( strncmp(arg,"-L",2) == 0 || strncmp(arg,"-l",2) == 0 )
			sprintf(libdir+strlen(libdir),"%s ",arg);
		else
		if( strncmp(arg,"-I",2) == 0 )
			sprintf(hdrdir+strlen(hdrdir),"%s ",arg);
	}

	tmp = tmpfile();
	if( !_available(tmp,"getcwd",DEFCC,hdrdir,libdir) )
		nav[nac++] = "-Dgetcwd=getwd";
	fclose(tmp);

	nav[nac] = NULL;

	EXECVP(nav[0],nav);
}

char *to_fullpath(path,fullpath)
	char *path,*fullpath;
{	char *dp,dc;
	char *path1,cwd[1024];

	GETCWD(fullpath,1024);
	if( (dp = strrchr(path,'/'))
	 || (dp = strrchr(path,'\\')) ){
		/* get the real base path into fullpath */
		path1 = dp + 1;
		strcpy(cwd,fullpath);
		dc = *dp;
		*dp = 0;
		chdir(path);
		*dp = dc;
		GETCWD(fullpath,1024);
		chdir(cwd);
	}else	path1 = path;
	strcat(fullpath,PATHSEP);
	if( strncmp(path1,"./",2) == 0 || strncmp(path1,".\\",2) == 0 )
		path1 += 2;
	strcat(fullpath,path1);
	return fullpath;
}

static baseof(path,base)
	char *path,*base;
{	char *bp,*dp;

	strcpy(base,path);
	dp = NULL;
	for( bp = base; *bp; bp++ )
		if( *bp == '/' || *bp == '\\' )
			dp = bp;
	if( dp )
		*dp = 0;
}

static delarg1(ac,av,ax)
	char *av[];
{	int ai;

	for( ai = ax; ai < ac; ai++ )
		av[ai] = av[ai+1];
	return ac - 1;
}

static char *COM;
static int flag_NEGATE;
static int flag_REUSE;

static scan_flags(flags)
	char *flags;
{	char *sp;

	for( sp = flags; *sp; sp++ ){
		switch( *sp ){
			case 'n': flag_NEGATE = 1; break;
			case 'r': flag_REUSE  = 1; break;
		}
	}
}

static putError(av,com,rcode)
	char *av[];
	char *com;
{	int ai;
	FILE *err;
	char cwd[128],path[1024];

	err = NULL;
	if( isatty(fileno(stdout)) )
	if( err = fopen("mkmake.err","a") ){
		GETCWD(path,sizeof(path));
		strcat(path,"/mkmake.err");
		fprintf(stderr,"mkmake: ERROR LOG is left at %s\n",path);
	}
	if( err == NULL )
		err = stdout;

	GETCWD(cwd,sizeof(cwd));
	fprintf(err,"\t*** exit(%d): %s at %s\n",rcode,com,cwd);
	for( ai = 0; av[ai]; ai++ )
		fprintf(err,"\t[%d] %s\n",ai,av[ai]);

	fprintf(err,"\n");
	if( err != stdout )
		fclose(err);
}

char *MYPATH;
char *MYBASE;

main(ac,av)
	char *av[];
{	int rcode;
	int ai;
	char mypath[1024],mybase[1024];

	setBinaryIO();

	if( strcmp(av[1],"-mkmkmk") == 0 )
		return mkmkmk(ac,av);

	COM = av[1];
	MYPATH = to_fullpath(av[0],mypath);
	baseof(MYPATH,mybase);
	MYBASE = mybase;

	if( 1 < ac && av[1][0] == '+' ){
		scan_flags(av[1]+1);
		ac = delarg1(ac,av,1);
	}

	if( strcmp(av[1],"-subst") == 0 ){
		ac = delarg1(ac,av,1);
		rcode = mksubst(ac,av);
	}else
	if( strcmp(av[1],"-lnconf") == 0 ){
		ac = delarg1(ac,av,1);
		rcode = lnconf(ac,av);
	}else
	if( strcmp(av[1],"-ckconf") == 0 ){
		ac = delarg1(ac,av,1);
		rcode = ckconf(ac,av);
	}else
	if( strcmp(av[1],"-makeat") == 0 ){
		ac = delarg1(ac,av,1);
		rcode = makeat(ac,av);
	}else
	if( strcmp(av[1],"-makeit") == 0 ){
		ac = delarg1(ac,av,1);
		rcode = makeit(ac,av);
	}else
	if( strcmp(av[1],"-lkfile") == 0 ){
		ac = delarg1(ac,av,1);
		rcode = lkfile(ac,av);
	}else
	if( strcmp(av[1],"-mkmake") == 0 ){
		ac = delarg1(ac,av,1);
		rcode = mkmake(ac,av);
	}else{
		char cwd[1024];
		GETCWD(cwd,sizeof(cwd));
		fprintf(stderr,"ERROR: unknown usage of mkmake: %s\n",cwd);
		for( ai = 0; ai < ac; ai++ )
			fprintf(stderr," [%d] %s\n",ai,av[ai]);
		exit(-1);
	}

	if( rcode != 0 )
		putError(av,COM,rcode);

	if( flag_NEGATE ){
		if( rcode != 0 )
			rcode = 0;
		else	rcode = 1;
	}
	return rcode;
}

/*
####  lnconf -- link DELEGATE_CONF at upper directories
#### 
*/
lnconf(ac,av)
	char *av[];
{	char rwd[1024],cwd[1024],pwd[1024],dir[1024];
	char opath[1024];
	char command[1024];
	char *CONF;
	char *LINK;
	int fd;
	FILE *fp;

/*
	printf("\tlnconf:\n");
*/

	CONF = av[1];
	if( 0 <= (fd = open(CONF,0)) ){
		close(fd);
		return 0;
	}

	GETCWD(rwd,sizeof(rwd));
	pwd[0] = 0;

	dir[0] = 0;
	for(;;){
		GETCWD(cwd,sizeof(cwd));
		if( fopen(CONF,"r") != NULL ){
			if( dir[0] == 0 )
				return 0;
			chdir(rwd);
			sprintf(opath,"%s%s%s",dir,PATHSEP,CONF);
#ifdef S_IFLNK
			if( unlink(CONF) == 0 )
			printf("\told symlink removed\n");
			printf("\tsymlink %s %s\n",opath,CONF);
			return symlink(opath,CONF);
#else
			printf("\tcopy %s %s\n",opath,CONF);
			return cpyfile(CONF,"w",opath,"*");
#endif
		}
		if( strcmp(cwd,pwd) == 0 )
			break;
		strcpy(pwd,cwd);

		if( chdir("..") == 0 )
			sprintf(dir+strlen(dir),"..%s",PATHSEP);
		else	break;
		printf("\t#### looking for %s%s\n",dir,CONF);
		fflush(stdout);
	}

	/* make template */
	chdir(rwd);
	if( fp = fopen(CONF,"w") ){
		GETCWD(cwd,sizeof(cwd));
		fprintf(stderr,"\tcreated %s%s%s\n",cwd,PATHSEP,CONF);
		fprintf(fp,"#### DELEGATE CONFIGURATION FOR MAKE ####\n");
		fclose(fp);
		return 0;
	}else{
		fprintf(stderr,"cannot create %s/%s\n",rwd,CONF);
		return -1;
	}
}

/*
####  mkmake -- make Makefile.go
####  mkmake MAKE OUT CFLAGS LDFLAGS RANLIB LIBLIST LIBNAME [LIBSRCS]
*/

char *MAKE;
char *OUT;
char *CFLAGS;
char *LDFLAGS;
char *RANLIB;
char *LIBLIST;
char *LIBNAME;
char *LIBSRCS;

char *CONF = "DELEGATE_CONF";
char *SRC = "Makefile";

#define _make	"_make"
#define _c	"_.c"
#define _out	"_.out"
#define _a	"_.a"

#define _ranlib	"ranlib"
static find_ranlib(ranlib,make)
	char *ranlib,*make;
{ 	char command[128],res[0x10000];
	int rcode;

	if( which(_ranlib,ranlib) )
		return 1;

	unlink(_a);
	putfile(_c,"w","sub(){}\n");
	putfile(_make,"w",
"_.a:; $(CC) -c _.c\n\t$(AR) cr _.a _.o\n\t$(AR) st _.a\n\techo SUCCESS\n");
	sprintf(command,"%s -f _make",make);
	rcode = msystem(1,res,sizeof(res),command);
	unlink(_c);
	unlink(_make);
	unlink(_a);
	unlink("_.o");

	if( rcode == 0 && STRSTR(res,"SUCCESS") ){
		strcpy(ranlib,"$(AR) st");
		return 1;
	}

	return 0;
}

#define strspace(s)	strpbrk(s," \t\r\n")

mkmake(ac,av)
	char *av[];
{	FILE *out;
	char ranlib[1024];
	int fd;
	char conf[1024],cwd[1024];
	char libsrcs[0x4000];

	printf("\tmkmake:\n");

	MAKE = av[1];
	OUT = av[2];
	CFLAGS = av[3];
	LDFLAGS = av[4];
	RANLIB = av[5];
	LIBLIST = av[6];
	LIBNAME = av[7];
	LIBSRCS = av[8];

	if( flag_REUSE )
	if( 0 <= (fd = open(OUT,0)) ){
		printf("\t#### %s already made.\n",OUT);
		close(fd);
		return 0;
	}

	libsrcs[0] = 0;
	if( LIBSRCS ){
		char cwd[1024],lscom[1024];
		char *sp,*lp,*fp,*dp;
		char name[128],dir1[128],files[4096],file1[128];
		char *env;

		/* to ignore -w flag of GNUmake ... */
		if( env = getenv("MAKEFLAGS") ) *env = 0;
		if( env = getenv("MFLAGS") ) *env = 0;

		GETCWD(cwd,sizeof(cwd));
		sprintf(lscom,"%s %s %s",MAKE,BESILENT,SRCFILE_LIST);
		dp = libsrcs;
		for( sp = LIBSRCS; sp; sp = strspace(sp+1) ){
			while( *sp && strchr(" \t\r\n",*sp) )
				sp++;
			if( sscanf(sp,"%[^=]=%s",name,dir1) != 2 )
				break;
			printf("\tscanning source of %s at %s\n",name,dir1);
			sprintf(dp,"LIBSRC_%s=",name);
			dp += strlen(dp);

			if( chdir(dir1) != 0 ){
				printf("*** No Such Directory ?\n");
				exit(-1);
			}

			files[0] = 0;
			msystem(0,files,sizeof(files),lscom);
			chdir(cwd);
			printf("%s",files);
			if( strstr(files,"Makefile") == 0 ){
				printf("*** No Makefile ?\n");
				exit(-1);
			}

			for( fp = files; fp; fp = strspace(fp+1) ){
				while( *fp && strchr(" \t\r\n",*fp) )
					fp++;
				if( *fp == 0 )
					break;

				file1[0] = 0;
				sscanf(fp,"%s",file1);
			if( file1[0] == 0 )
					continue;

				if( strchr(file1,'.') == 0 )
				if( strcmp(file1,"Makefile") != 0 )
				{
					printf("#### won't use [%s]\n",file1);
					continue;
				}

				if( strcmp(file1,"\"") == 0 )
					continue; /* for Win95/98 echo */

				if( dp != libsrcs ){
					sprintf(dp," \\\n ");
					dp += strlen(dp);
				}
				sprintf(dp,"%s%s%s",dir1,PATHSEP,file1);
				dp += strlen(dp);
			}
			sprintf(dp,"\n");
			dp += strlen(dp);
		}
	}

	printf("\t#### creating %s\n",OUT);
	out = fopen(OUT,"w");
	fprintf(out,"#### Generated from SRC=%s and CONF=%s\n",SRC,CONF);
	fflush(out);

	cpyfile(OUT,"a",SRC,"1","/#---BGN---",1);
	fseek(out,0,2);

	printf("\t[SET] MKMAKE=%s\n",MYPATH);
	fprintf(out,"MKMAKE=%s\n",MYPATH);
	fprintf(out,"MKBASE=%s\n",MYBASE);
	fprintf(out,"MKMKMK=%s%s%s\n",MYBASE,PATHSEP,MKMKMK_EXE);
	fprintf(out,"CFLAGS=%s\n",CFLAGS);
	fprintf(out,"LDFLAGS=%s\n",LDFLAGS);
	fprintf(out,"%s\n",libsrcs);

	if( !find_ranlib(ranlib,MAKE) ){
		strcpy(ranlib,RANLIB);
		if( ranlib[0] == 0 )
			strcpy(ranlib,"echo");
	}
	fprintf(out,"RANLIB=%s\n",ranlib);
	printf("\t[SET] RANLIB=%s\n",ranlib);

	fflush(out);
	cklibs(out,MAKE);

	fprintf(out,"#---CONF=%s\n",CONF);
	fflush(out);
	cpyfile(OUT,"a",CONF,"1","$",1);
	fseek(out,0,2);

	cpyfile(OUT,"a",SRC,"/#---END---","$",1);
	fseek(out,0,2);
	return 0;
}

/*
####  cklibs -- Pickup available libraries then print it to $OUT
####            from given library-list $LIBLIST
*/
cklibs(out,make)
	FILE *out;
	char *make;
{	char SLIBLIST[1024];
	char lv[256][128],*l1;
	int lc,li;
	char line[1024];
	char command[1024];
	int rcode;
	FILE *pfp;
	char res[0x10000];

	putfile(_c,"w","main(){}\n");
	/*
	lc = sscanf(LIBLIST,"%s %s %s %s %s %s %s %s",
		lv[0],lv[1],lv[2],lv[3],lv[4],lv[5],lv[6],lv[7]);
	*/
	lc = sscanf(LIBLIST,"%s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
		lv[0],lv[1],lv[2],lv[3],lv[4],lv[5],lv[6],lv[7],
		lv[8],lv[9],lv[10],lv[11],lv[12],lv[13]);

	SLIBLIST[0] = 0;
	unlink("_.out");

	for( li = 0; li < lc; li++ ){
		char *lp;

		l1 = lv[li];
		printf("\t- checking availability of %s [%s]\n",l1,SLIBLIST);
		lp = &SLIBLIST[strlen(SLIBLIST)];
		if( lp == SLIBLIST )
			strcpy(lp,l1);
		else	sprintf(lp," %s",l1);

		cpyfile(_make,"w",CONF,"1","$",1);
		sprintf(line,"\n_.out:; $(CC) %s%s %s %s\n\techo SUCCESS\n",
			CCOUT,_out,_c,SLIBLIST);
		putfile(_make,"a",line);

		sprintf(command,"%s -f _make",make);
		rcode = msystem(1,res,sizeof(res),command);
		if( STRSTR(res,"SUCCESS") == NULL )
			rcode = -1;

		if( unlink("_.out") != 0 )
			if( rcode == 0 )
				rcode = -1;
		if( rcode != 0 ){
			*lp = 0;
			printf("\t*** %s exit(%d)\n",COM,rcode);
		}
	}

	unlink("_make");
	unlink("_.c");
	unlink("_.out");

	printf("\t[SET] %s=%s\n",LIBNAME,SLIBLIST);
	fprintf(out,"%s=%s\n",LIBNAME,SLIBLIST);
	fflush(out);
}

/*
####  ckconf -- check configuration
####
*/
ckconf(ac,av)
	char *av[];
{	int ai;
	char *user,hostname[256],admin[256],ans[256],*dp;
	char line[1024];
	FILE *conf,*okfp;
	int fd;
	FILE *tty;

	char *MAKEFILE_OK;
	char *MAKEFILE;
	char *CONF;
	char *ADMIN;
	char *AR;

	printf("\tckconf:\n");

	if( ac < 5 ){
		fprintf(stderr,
		"Usage: %s MAKEFILE_OUT MAKEFILE ADMIN\n",
		av[0]);
		return 1;
	}

	MAKEFILE_OK = av[1];

	if( flag_REUSE )
	if( 0 <= (fd = open(MAKEFILE_OK,0)) ){
		printf("\t#### %s already made.\n",MAKEFILE_OK);
		close(fd);
		return 0;
	}

	MAKEFILE = av[2];
	CONF = av[3];
	AR = av[4];
	ADMIN = av[5];

	printf("\tADMIN = '%s'\n",ADMIN);
	if( *ADMIN == 0 || strchr(ADMIN,'@') )
		goto conf_ok;

	if( isatty(fileno(stderr)) )
		tty = stderr;
	else{
		tty = fopen("/dev/tty","w");
		if( tty == NULL ){
			fprintf(stderr,"## can't write to tty\n");
			return 1;
		}
	}

printf("\tget ADMIN value interactively ...\n");
fprintf(tty,"!!!!!!!!!!!!!!!!!!!! WARNING !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
fprintf(tty,"!   You should have defined ADMIN as:                      !\n");
fprintf(tty,"!     ADMIN = mail-address-of-the-DeleGate-administrator   !\n");
fprintf(tty,"!   either in %s or in the Makefile.      !\n",CONF);
fprintf(tty,"!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");

	sysgethostname(hostname);
	if( (user=getenv("USER")) || (user=getenv("LOGNAME")) )
		sprintf(admin,"%s@%s",user,hostname);
	else	admin[0] = 0;

	for(;;){

fprintf(tty,"Enter your E-mail address [%s]: ",admin);
fflush(tty);

		if( fgets(ans,sizeof(ans),stdin) == NULL )
			return 1;
		if( dp = strpbrk(ans,"\r\n") )
			*dp = 0;
		if( ans[0] != 0 )
			strcpy(admin,ans);

fprintf(tty,"ADMIN = %s\n",admin);
fprintf(tty,"OK ?  [y] / n / x(abort): ");
fflush(tty);

		if( fgets(ans,sizeof(ans),stdin) == NULL )
			return 1;
		if( dp = strpbrk(ans,"\r\n") )
			*dp = 0;

		if( ans[0] == 'y' || ans[0] == 0 )
			break;

		if( ans[0] == 'n' )
			continue;

		if( ans[0] == 'x' )
			return 1;

		fprintf(tty,"Answer y / n / RETURN\n");
	}

	conf = fopen(CONF,"a");
	if( conf == NULL ){
		fprintf(tty,"\n");
		return 1;
	}
	printf("\tgot ADMIN = %s interactively\n",admin);
	fprintf(conf,"#### ADMIN value got interactively ####\n");
	fprintf(conf,"ADMIN=%s\n",admin);
	fclose(conf);

conf_ok:
	if( *AR == 0 ){
		if( conf = fopen(CONF,"a") ){
			fprintf(conf,"#### AR added automatically ####\n");
			fprintf(conf,"AR=ar\n");
			fclose(conf);
		}
	}
	if( okfp = fopen(MAKEFILE_OK,"w") ){
		fprintf(okfp,"ckconf ok: %s + %s\n",MAKEFILE,CONF);
		fclose(okfp);
		return 0;
	}
	fprintf(stderr,"ckconf: cannot create %s\n",MAKEFILE_OK);
	return 1;
}

/*
##  makeit MAKE ...
##  0      1
*/
makeit(ac,av)
	char *av[];
{	char make[1024],makeenv[1024];
	int ai;
	char **nav;

	if( ac < 2 ){
		fprintf(stderr,"Usage: %s MAKE ...\n",av[0]);
		exit(1);
	}

	for( ai = 0; av[ai]; ai++ )
		if( strncmp(av[ai],"MAKE=",5) == 0 )
			av[ai] = MAKEenv(av[ai],makeenv);

	if( av[1][0] == 0 ){
		av[0] = "make";
		sprintf(make,"MAKE=%s",av[0]);
		av[1] = make;
		nav = &av[0];
	}else{
		sscanf(av[1],"%s",make);
		av[1] = make;
		nav = &av[1];
	}
	EXECVP(nav[0],nav);
	fprintf(stderr,"cannot call make: %s",nav[0]);
	exit(-1);
}


/*
##  makeat To Work Out MAKE Target Remain
##  0      1  2    3   4    5
*/

extern char *addArg();
makeat(ac,av)
	char *av[];
{	char copy[1024];
	char cwd[1024],rwd[1024],dst[1024],src[1024],*pp,*cpycom;
	int rcode;
	char mkmake[1024],mkbase[1024],mkmkmk[1024],makeenv[1024];
	int ai,nac;
	char *nav[1024];

	if( ac < 5 ){
		fprintf(stderr,"Usage: %s TARGET WHERE OUTFILE MAKE ...\n",
		av[0]);
		return 1;
	}

	strcpy(dst,av[1]);
	GETCWD(cwd,sizeof(cwd));

	if( chdir(av[2]) != 0 ){
		fprintf(stderr,"### cannot chdir to %s\n",av[2]);
		return 1;
	}
	GETCWD(rwd,sizeof(rwd));

	if( 5 < ac )
	if( strcmp(av[5],"=") == 0 )
		av[5] = av[3];

	nac = ac - 4;
	nav[0] = av[4];
	for( ai = 1; ai < nac; ai++ )
		nav[ai] = av[ai+4];
	sprintf(mkmake,"MKMAKE=%s",MYPATH);
	nav[nac++] = mkmake;
	sprintf(mkbase,"MKBASE=%s",MYBASE);
	nav[nac++] = mkbase;
	sprintf(mkmkmk,"MKMKMK=%s%s%s",MYBASE,PATHSEP,MKMKMK_EXE);
	nav[nac++] = mkmkmk;

	nav[nac] = NULL;

	if( nav[0][0] == 0 ){
		nav[0] = "make";
		nav[nac++] = MAKEenv("MAKE=make",makeenv);
		nav[nac] = NULL;
	}
	for( ai = 0; nav[ai]; ai++ )
		if( strncmp(nav[ai],"MAKE=",5) == 0 )
			nav[ai] = MAKEenv(nav[ai],makeenv);

	printf("\r\n####### %s: %s\n",COM,rwd);
	fflush(stdout);

	if( rcode = EXECVPR(nav[0],nav) ){
		putError(nav,"-makeat",rcode);
		return rcode;
	}

	if( av[1][0] == 0 )
		return 0;

	if( chdir(cwd) != 0 ){
		printf("cannot chdir to: %s\n",cwd);
		return 1;
	}

	if( av[3][0] == '/' )
		strcpy(src,av[3]);
	else	sprintf(src,"%s/%s",av[2],av[3]);

	if( (rcode = renames(src,dst)) == 0 )
		cpycom = "move";
	else
	if( (rcode = cpyfile(dst,"w",src,"*")) == 0 )
		cpycom = "copy";
	else	cpycom = "move/copy";

	if( rcode == 0 )
		unlink(src);

	printf("**[%d] %s %s to %s\n",rcode,cpycom,src,dst);
	return rcode;
}


/*
##  lkfile
## 
*/
lkfile(ac,av)
	char *av[];
{	char *src,*dst,*dp;

	if( ac < 3 ){
		printf("\tUsage: lkfile file-existing file-newname\n");
		return 1;
	}
	src = av[1];
	dst = av[2];
	if( strcmp(dst,".") == 0 ){
		if( dp = strrchr(src,'/') ){
			dst = dp + 1;
			printf("\tlkfile %s %s\n",src,dst);
		}else{
			printf("\tlkfile: no destination file\n");
			return 1;
		}
	}
	return cpyfile(dst,"w",src,"*");
}

/*
####  library functions
####
*/
cpyfile(out,mode,in,from,to,useifdef)
	char *out,*mode,*in,*from,*to;
{	FILE *ofp,*ifp;
	char line[1024];
	char *ls1,*ls2;
	int ln1,ln2,lnc;
	int putit;
	int nput;
	int ignore;
	int rcode;

	if( (ifp = fopen(in,"r")) == NULL ){
		printf("\tcpyfile: cannot read '%s'\n",in);
		return 1;
	}

	if( (ofp = fopen(out,mode)) == NULL ){
		printf("\tcpyfile: cannot write '%s'\n",out);
		fclose(ifp);
		return 1;
	}

	rcode = -1;
	if( strcmp(from,"*") == 0 ){
		int rcc,wcc;
		char buf[1024];
		while( rcc = fread(buf,1,sizeof(buf),ifp) )
			wcc = fwrite(buf,1,rcc,ofp);
		rcode = 0;
		goto EXIT;
	}

	if( *from == '/' )
		ls1 = from+1;
	else	ls1 = NULL;

	if( *to == '/' )
		ls2 = to+1;
	else	ls2 = NULL;

	ln1 = atoi(from);
	ln2 = atoi(to);

	putit = 0;
	ignore = 0;
	nput = 0;

	printf("\tcpyfile %s[%d-%d][%x-%x] > %s\n",in,ln1,ln2,ls1,ls2,out);

	for( lnc = 1; fgets(line,sizeof(line),ifp) != NULL; lnc++ ){
		if( !putit )
		if( ln1 && ln1 <= lnc
		 || ls1 && strncmp(line,ls1,strlen(ls1)) == 0 )
			putit = 1;

		if( !ignore && useifdef )
		if( strncmp(line,"#ifdef",6) == 0 ){
			if( !matchList(line+7,SYST) )
				ignore = 1;
			else	printf("\t%s",line);
		}

		if( putit ){
			if( ignore )
				fputs("## ",ofp);
			else	nput++;
			fputs(line,ofp);
		}

		if( ignore && useifdef )
		if( strncmp(line,"#endif",6) == 0 ){
			ignore = 0;
		}

		if( putit )
		if( ln2 && ln2 <= lnc
		 || ls2 && strncmp(line,ls2,strlen(ls2)) == 0 ){
			putit = 0;
			break;
		}
	}
EXIT:
	fclose(ifp);
	fclose(ofp);
	return rcode;
}
matchList(list,word)
	char *list,*word;
{	char *wp,*ep,ec,listbuf[1024];

	strcpy(listbuf,list);
	wp = listbuf;
	for( ep = wp; ec = *ep; ep++ ){
		if( strchr(", #\t\r\n",ec) ){
			*ep = 0;
			if( strcmp(word,wp) == 0 )
				return 1;
			wp = ep + 1;
		}
	}
	return 0;
}

putfile(out,mode,str)
	char *out,*mode,*str;
{	FILE *ofp;

	if( (ofp = fopen(out,mode)) == NULL ){
		printf("\tputfile: cannot write '%s'\n",out);
		return 1;
	}

	fputs(str,ofp);
	fclose(ofp);
	return 0;
}

which(command,path)
	char *command,*path;
{	char *pe,PATH[4096],*np,*dir;
	FILE *fp;

	if( pe = getenv("PATH") )
		strcpy(PATH,pe);
	else	PATH[0] = 0;
	strcat(PATH,":/bin:/usr/bin:/usr/ucb:/usr/local/bin");

	for( dir = PATH; dir && *dir; dir = np ){
		if( np = strchr(dir+1,':') )
			*np++ = 0;

		sprintf(path,"%s/%s",dir,command);
		if( fp = fopen(path,"r") ){
			fclose(fp);
			return 1;
		}
	}
	return 0;
}

renames(old,new)
	char *old,*new;
{	int ri,rcode;

	rcode = -1;
	unlink(new);

	for( ri = 0; ri < 5; ri++ ){
		if( (rcode = rename(old,new)) == 0 )
			break;
#ifdef RETRY_RENAME
			fprintf(stderr,"FAILED rename %s to %s, retry...\n",
				old,new);
			sleep(1); /* NMAKE problem ? */
#endif
	}
	if( rcode != 0 )
		fprintf(stderr,"FAILED rename(%s,%s)\n",old,new);
	return rcode;
}


/*
 *	970104 C program version of mksubst
 *		<ysato@etl.go.jp>
 */

#define IFEXIST "__"
#define TRY	"_-"

FILE *ERRLOG;
char *SRCDIR = "../src";

/*
char *MAKE;
*/
char *MAKEFILE;
char *CC;
char *HDRDIR;
/*
char *RANLIB;
*/
char *LIBFILE;
char *LIBDIR;
char *LDLIBS;

mksubst(ac,av)
	char *av[];
{	char *com;
	char *substfiles;

	ERRLOG = fopen("errors","a");
	if( ERRLOG == NULL )
		ERRLOG = stderr;

	if( ac < 6 ){
	fprintf(stderr,
"Usage: %s {libck|libmk|manmk} MAKE MAKEFILE CC CFLAGS HDRDIR RANLIB LIBFILE LIBDIR LDLIBS files\n",
	av[0]);
	exit(1);
	}

	com = av[1];
	MAKE = av[2];  if( MAKE[0] == 0 ) MAKE = "make";
	MAKEFILE = av[3];
	CC = av[4];    if( CC[0] == 0 ) CC = "cc";
	CFLAGS = av[5];
	HDRDIR = av[6];
	RANLIB = av[7];
	LIBFILE = av[8];
	LIBDIR = av[9];
	LDLIBS = av[10];
	substfiles = av[11];

	printf("\tMAKE=%s\n",MAKE);
	printf("\tMAKEFILE=%s\n",MAKEFILE);
	printf("\tCC=%s\n",CC);
	printf("\tCFLAGS=%s\n",CFLAGS);
	printf("\tHDRDIR=%s\n",HDRDIR);
	printf("\tRANLIB=%s\n",RANLIB);
	printf("\tLIBFILE=%s\n",LIBFILE);
	printf("\tLIBDIR=%s\n",LIBDIR);
	printf("\tLDLIBS=%s\n",LDLIBS);
	printf("\tFiles=%s\n",substfiles);

	if( strcmp(com,"libck") == 0 )
		exit(check_env());

	if( strcmp(com,"objmk") == 0 )
		exit(_mksubst(substfiles,NULL));

	if( strcmp(com,"libmk") == 0 )
		exit(libmake(substfiles,av[12],av[13]));

	if( strcmp(com,"manck") == 0 )
		exit(manifest_check(substfiles));

	if( strcmp(com,"manmk") == 0 )
		exit(manifest_make(substfiles));

	fprintf(stderr,"%s: unknown command [%s]\n",com);
	exit(1);
}

static exist1(file,loadit)
	char *file,*loadit;
{	FILE *fp;
	char *dp;

	if( dp = strrchr(file,'.') )
		strcpy(dp,".o");

	fp = fopen(file,"r");

	if( fp ){
		fclose(fp);
		if( loadit[0] != 0 )
			strcat(loadit," ");
		strcat(loadit,file);
	}
	return 0;
}

libmake(files,target,listname)
	char *files,*target,*listname;
{	char loadit[1024];
	char a1[1024];
	char *av[3];
	char *srcv[1024],*src1;
	int si;
	int rcode;

	srcv[0] = 0;
	_mksubst(files,srcv);
	loadit[0] = 0;
/*
	foreach_word(files,exist1,loadit);
*/
	for( si = 0; src1 = srcv[si]; si++ )
		exist1(src1,loadit);

	sprintf(a1,"%s=%s",listname,loadit);
	av[0] = a1;
	av[1] = target;
	av[2] = NULL;
	execMake("libmk",av,0);
}

manifest_check(files)
	char *files;
{	FILE *fp;
	char omani[1024],nenv[1024],nmani[4096];
	char manifest[1024];
	int rcode;
	char *av[2];

	sprintf(manifest,"mani-%s",LIBFILE);

sprintf(nenv,   "%s CC='%s' CFLAGS='%s' RANLIB='%s' LDLIBS='%s'",
MAKE,CC,CFLAGS,RANLIB,LDLIBS);
	sprintf(nmani,"%s %s",nenv,files);

	omani[0] = 0;
	if( fp = fopen(manifest,"r") ){
		fgets(omani,sizeof(omani),fp);
		fclose(fp);
		if( strcmp(omani,nmani) == 0 )
			return 0;
	}

	fprintf(stderr,"** manifest of '%s' changed:\n < %s\n > %s\n",
		LIBFILE,omani,nmani);

	unlink(LIBFILE);
	unlink(manifest);

	av[0] = LIBFILE;
	av[1] = NULL;
	execMake("manmk",av,0);
}
manifest_make(files)
	char *files;
{	FILE *fp;
	char nmani[4096];
	char manifest[1024];
	char *dp;

	sprintf(manifest,"%s.mani",LIBFILE);

sprintf(nmani,  "%s CC='%s' CFLAGS='%s' RANLIB='%s' LDLIBS='%s' %s",
MAKE,CC,CFLAGS,RANLIB,LDLIBS,files);

	if( fp = fopen(manifest,"w") ){
		fputs(nmani,fp);
		fclose(fp);
		return 0;
	}
	fprintf(stderr,"#### CANNOT MAKE tar MANIFEST %s\n",manifest);
	return 1;
}

#define stralloc(s)	strcpy((char*)malloc(strlen(s)+1),s)

subst1(file,filev,filecp)
	char *file;
	char *filev[];
	int *filecp;
{	char *dp;
	char symb[256],*sym,filebuf[256];
	int ifavail,try,avail;
	char libs[1024];

	dp = strrchr(file,'.');
	if( dp == NULL || strcmp(dp,".c") != 0 )
		return 0;

	strcpy(symb,file);
	*strrchr(symb,'.') = 0;

	sym = symb;
	ifavail = 0;
	try = 0;
	if( strncmp(sym,IFEXIST,strlen(IFEXIST)) == 0 ){
		sym += strlen(IFEXIST);
		ifavail = 1;
	}else
	if( strncmp(sym,TRY,strlen(TRY)) == 0 ){
		sym += strlen(TRY);
		ifavail = 1;
		try = 1;
	}

	sprintf(libs,"%s %s",LIBDIR,LDLIBS);
	avail = _available(ERRLOG,sym,CC,CFLAGS,libs);

	if(  ifavail ){
		if( avail )
			fprintf(stderr," ** %s.c is used\n",symb);
		else
		if( try )
fprintf(stderr," -- %s may not be in the libraries, but try to get it.\n",symb);
		else	fprintf(stderr," -- %s is not in the libraries\n",symb);

		if( avail || try ){
			sprintf(filebuf,"%s.o",symb);
			filev[*filecp] = stralloc(filebuf);
			*filecp = *filecp + 1;
		}
	}else{
		if( !avail ){
			fprintf(stderr," ** %s.c is substituted\n",symb);
			sprintf(filebuf,"%s.o",symb);
			filev[*filecp] = stralloc(filebuf);
			*filecp = *filecp + 1;
		}else	fprintf(stderr," -- %s is in the libraries\n",symb);
	}
	return 0;
}

_mksubst(files,filev)
	char *files;
	char *filev[];
{	char cwd[1024];
	char *filevb[128];
	int filec;

	fprintf(stderr,"** mksubst **\n");
	fprintf(stderr,"  where: %s\n",SRCDIR);
	fprintf(stderr,"  compile: %s %s %s\n",CC,CFLAGS,LDLIBS);
	fprintf(stderr,"  files: %s\n",files);

	/*
	 * Goto SRCDIR because the CFLAGS,LDLIBS may be specified in
	 * relative path from the directory
	 */
	GETCWD(cwd,sizeof(cwd));
	if( chdir(SRCDIR) != 0 ){
		fprintf(stderr,"cannot chdir SRCDIR = %s\n",SRCDIR);
		return -1;
	}

	setSIGINT("mksubst");

	filec = 0;
	if( filev == NULL )
		filev = filevb;
	foreach_word(files,subst1,filev,&filec);
	chdir(cwd);

	filev[filec] = NULL;
	if( filev == filevb )
		execMake("mksub",filev,0); /* no return */
	else	return callMake("mksub",filev);
}

fatal(fmt,a,b,c,d,e,f,g)
	char *fmt,*a,*b,*c,*d,*f,*g;
{
	fprintf(stderr,"FATAL!!!!:");
	fprintf(stderr,fmt,a,b,c,d,e,f,g);
	fprintf(stderr,"\n");
}

check_env(){
	char libs[1024];

	fprintf(stderr,"\t -- checking CC = %s\n",CC);
	if( !_available(ERRLOG,NULL,CC,"",LIBDIR) ){
		fatal(" Something wrong in the CC.");
		return 1;
	}
	fprintf(stderr,"\t -- checking CFLAGS = %s\n",CFLAGS);
	if( !_available(ERRLOG,NULL,CC,CFLAGS,LIBDIR) ){
		fatal(" Something wrong in the CFLAGS.");
		return 2;
	}
	sprintf(libs,"%s %s",LIBDIR,LDLIBS);
	fprintf(stderr,"\t -- checking LDLIBS = %s\n",LDLIBS);
	if( !_available(ERRLOG,NULL,CC,CFLAGS,libs) ){
		fatal(" Something wrong in Libraries.");
		return 3;
	}

	if( MaybeOnUnix ){
		fprintf(stderr,"\t -- checking socket library\n");
		if( !_available(ERRLOG,"socket",CC,CFLAGS,libs) ){
#ifndef __CYGWIN__
			fatal(" Socket library should be given.");
			fatal(" You may have to specify");
			fatal("     LIBS=-lnsl -lsocket");
			fatal("     LIBS=WSOCK32.LIB");
			fatal("     ...");
			fatal(" in the Makefile (or in DELEGATE_CONF)");
			return 4;
#endif
		}
	}

	fprintf(stderr,"** CC CFLAGS LIBS and socket library are OK.\n");
	return 0;
}

foreach_word(list,func,a1,a2,a3)
	char *list;
	int (*func)();
	char *a1,*a2,*a3;
{	char *np;
	char file[1024];
	int rcode;

	for( np = list; np; np = strpbrk(np," \t") ){
		while( *np == ' ' || *np == '\t' )
			np++;
		if( *np == 0 )
			break;
		sscanf(np,"%s",file);
		if( rcode = (*func)(file,a1,a2,a3) )
			break;
	}
	return rcode;
}

callMake(what,aav)
	char *what;
	char *aav[];
{	int pid;
 
#ifdef WITH_SPAWN
	return execMake(what,aav,1);
#else
	if( (pid = fork()) == 0 ){
		execMake(what,aav,0);
		exit(-1);
	}
	return waitPid(pid);
#endif
}

extern char *addArg();
execMake(what,aav,spawn)
	char *what;
	char *aav[];
{	char *av[256],ab[2048],*bp;
	int ac = 0, ai;
	int aac;

	ac = 0;
	bp = ab;

	av[ac++] = MAKE;
	av[ac++] = addArg(&bp,"MAKE",MAKE);
	if( MAKEFILE[0] != 0 ){
	av[ac++] = "-f";
	av[ac++] = addArg(&bp,MAKEFILE,NULL);
	}
	av[ac++] = addArg(&bp,"CC",CC);
	av[ac++] = addArg(&bp,"CFLAGS",CFLAGS);
	av[ac++] = addArg(&bp,"HDRDIR",HDRDIR);
	av[ac++] = addArg(&bp,"RANLIB",RANLIB);
	av[ac++] = addArg(&bp,"LIBFILE",LIBFILE);

	for( aac = 0; aav[aac]; aac++ )
		av[ac++] = addArg(&bp,aav[aac],NULL);
	av[ac++] = NULL;

	for( ai = 0; av[ai]; ai++ )
		printf("\t%s[%d] %s\n",what,ai,av[ai]);

#ifdef WITH_SPAWN
	if( spawn ){
		int rcode;
		if( rcode = spawnvp(P_WAIT,MAKE,av) )
			printf("*** mksubst exit(%d) %s\n",av[0]);
		return rcode;
	}
#endif
	EXECVP(MAKE,av);

	fprintf(stderr,"EXECVP(%s) failed\n",MAKE);
	perror("EXECVP");
	exit(1);
}

/*
 *	970204 C program version of available
 *		<ysato@etl.go.jp>
 */
#include <signal.h>

#define TEST_C	"_.c"

setSIGINT(what)
	char *what;
{
	signal(SIGINT,SIG_DFL);
}

static isfile(path)
	char *path;
{	struct stat st;

	return stat(path,&st) == 0;
}

_available(errlog,sym,cc,flags,libs)
	FILE *errlog;
	char *sym,*cc,*flags,*libs;
{	FILE *fp;
	int rcode,fd;
	char command[1024],res[0x10000];

	fp = fopen(TEST_C,"w");
	if( sym != NULL )
		fprintf(fp,"main(){ %s();}\n",sym);
	else{
		sym = "ANYTHING";
		fprintf(fp,"main(){ }\n");
	}
	fclose(fp);
	unlink("a.out");
	unlink("_.exe");
	unlink("a.exe");

	sprintf(command,"%s %s %s %s",cc,flags,TEST_C,libs);
	rcode = msystem(1,res,sizeof(res),command);

	if( errlog != NULL ){
		fprintf(errlog,"***\n%s\n",command);
		fprintf(errlog,"%s\n",res);
		fflush(errlog);
	}else	printf("%s",res);

	if( !isfile("a.out") && !isfile("_.exe") && !isfile("a.exe") ){
		if( strstr(res,"too few arguments") ){
			fprintf(stderr,"*** %s : available ?\r\n",sym);
			rcode = 0;
		}else
		if( rcode == 0 ){
			fprintf(stderr,"*** %s : unavailable ?\r\n",sym);
			rcode = -1;
		}
	}

	unlink("a.out");
	unlink("_.exe");
	unlink("a.exe");

	if( rcode == 0 )
		unlink(TEST_C);

	return rcode == 0;
}

char *addArg(pp,name,value)
	char **pp,*name,*value;
{	char *p0,arg[1024];

	p0 = *pp;
	if( value == NULL )
		strcpy(arg,name);
	else	sprintf(arg,"%s=%s",name,value);

#ifdef QUOTE_ARG
	if( arg[0] != '"' )
		sprintf(p0,"\"%s\"",arg);
	else
#endif
	strcpy(p0,arg);

	*pp = p0 + strlen(p0) + 1;
	return p0;
}


EXECVPR(path,av)
	char *path,*av[];
{	char *nav[256],na[4][256];
	char *nap,nab[1024];
	int nac0,ai,nac;
	int status;
	int pid,rcode;

	if( strpbrk(path," \t") ){
		na[0][0] = 0;
	 	nac0 = sscanf(path,"%s %s %s %s",na[0],na[1],na[2],na[3]);
		fprintf(stderr,"#### execvp(%s) -> execpv(%s)\n",
			av[0],na[0]);
		av[0] = na[0];
		path = av[0];
/*
some make produeces MAKE="make TARGET=x\ y\ z" ...
		path = na[0];
		nac = 0;
		fprintf(stderr,"#### execvp(%s):\n",av[0]);
		for(ai = 0; ai < nac0; ai++){
			printf("\t[%d] %s\n",ai,na[ai]);
			nav[nac++] = na[ai];
		}
		for(ai = 1; av[ai]; ai++ )
			nav[nac++] = av[ai];
		nav[nac] = NULL;
		av = nav;
*/
	}

#ifdef WITH_SPAWN
	nav[0] = av[0];
	nap = nab;
	for( ai = 1; av[ai]; ai++ )
		nav[ai] = addArg(&nap,av[ai],NULL);
	nav[ai] = 0;
	rcode = spawnvp(P_WAIT,nav[0],nav);
#else
	if( (pid = fork()) == 0 ){
		rcode = Execvp(av[0],av);
		fprintf(stderr,"#### execvp(%s) = %d, failed\n",path,rcode);
		exit(-1);
	}else{
		rcode = waitPid(pid);
	}
#endif
	return rcode;
}

#ifndef WITH_SPAWN
waitPid(pid)
{	int xpid,status,rcode;

	rcode = -1;
	for(;;){
		xpid = wait(&status);
		if( xpid != -1 )
			rcode = WEXITSTATUS(status);
		else	rcode = -1;
		/*printf("waitPid(%d) = %d, exit(%d)\n",pid,xpid,rcode);*/
		if( xpid == -1 || xpid == pid )
			break;
	}
	return rcode;
}
#endif

EXECVP(path,av)
	char *path,*av[];
{
	exit(EXECVPR(path,av));
}

msystem(erralso,res,size,command)
	char *res,*command;
{	int rcc,ecc;
	FILE *out,*errout;
	int outpipe[2];
	int outfd,errfd;
	int rcode;

	pipe(outpipe);
	errout = tmpfile();

	outfd = dup(1);
	dup2(outpipe[1],1);
	if( erralso ){
		errfd = dup(2);
		dup2(fileno(errout),2);
	}
	rcode = system(command);
	dup2(outfd,1);
	close(outfd);
	if( erralso ){
		dup2(errfd,2);
		close(errfd);
	}

	close(outpipe[1]);
	out = fdopen(outpipe[0],"r");
	rcc = fread(res,1,size,out);
	fclose(out);

	fseek(errout,0,0);
	ecc = fread(res+rcc,1,size,errout);
	fclose(errout);
	res[rcc+ecc] = 0;
	return rcode;
}

#ifndef HOSTNAME
#define HOSTNAME	NULL
#endif
#ifndef HOSTNAME_FILE
#define HOSTNAME_FILE	NULL
#endif
#ifndef HOSTNAME_COM
#define HOSTNAME_COM	"hostname"
#endif

sysgethostname(hostname)
	char *hostname;
{	FILE *fp;

	*hostname = 0;

	if( HOSTNAME ){
		strcpy(hostname,HOSTNAME);
		return 1;
	}

	if( HOSTNAME_FILE ){
		if( fp = fopen(HOSTNAME_FILE,"r") ){
			fscanf(fp,"%s",hostname);
			fclose(fp);
			if( *hostname )
				return 1;
		}
	}
	if( HOSTNAME_COM ){
		if( fp = popen(HOSTNAME_COM,"r") ){
			fscanf(fp,"%s",hostname);
			pclose(fp);
			if( *hostname )
				return 1;
		}
	}
	return 0;
}

#ifdef WITH_SPAWN
Execvp(path,argv)
	char *path,*argv[];
{	int stat;

	stat = spawnvp(P_WAIT,path,argv);
	if( stat == -1 )
		return -1;
	else	exit(stat);
}
#else
Execvp(path,argv)
	char *path,*argv[];
{
	return execvp(path,argv);
}
#endif
